/***********************************************************************

This file is a simple script written in C to apply the Page's
trend test to analyze evolutionary search algorithms convergence.
Results are obtained in LaTeX and plain text formats.

Copyright (C) 2014
	
J. Derrac (derrac@decsai.ugr.es)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see http://www.gnu.org/licenses/
  
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>

int nAlgs;
int nFunc;
int nCuts;

char ** names = NULL;

#define UNDEFINED 999
#define MAX_ALGS 20
#define MAX_FUNC 100
#define MAX_CUTS 1000

#define LARGE_DOUBLE 999999.9

double results [MAX_ALGS][MAX_FUNC][MAX_CUTS];
double differences [MAX_FUNC][MAX_CUTS];

double ranks [MAX_ALGS][MAX_ALGS][MAX_FUNC][MAX_CUTS];
double sumRanks [MAX_ALGS][MAX_ALGS][MAX_CUTS];
double L [MAX_ALGS][MAX_ALGS];
double pValue [MAX_ALGS][MAX_ALGS];

double sumRanksCorrected [MAX_ALGS][MAX_ALGS][MAX_CUTS];
double LCorrected [MAX_ALGS][MAX_ALGS];
double pValueCorrected [MAX_ALGS][MAX_ALGS];

char *record = NULL;
char *line = NULL;

char buffer[1024];
    
void computeDifferences(int algA, int algB){
		
	int i,j;

	for(i=0;i<nFunc;i++){		
		for(j=0;j<nCuts;j++){
			differences[i][j]=results[algA][i][j]-results[algB][i][j];
		}
	}
}

double getTipifiedProbability(double z, int upper) {
        
	double ltone = 7.0,
	utzero = 18.66,
	con = 1.28,
	a1 = 0.398942280444,
	a2 = 0.399903438504,
	a3 = 5.75885480458,
	a4 = 29.8213557808,
	a5 = 2.62433121679,
	a6 = 48.6959930692,
	a7 = 5.92885724438,
	b1 = 0.398942280385,
	b2 = 3.8052e-8,
	b3 = 1.00000615302,
	b4 = 3.98064794e-4,
	b5 = 1.986153813664,
	b6 = 0.151679116635,
	b7 = 5.29330324926,
	b8 = 4.8385912808,
	b9 = 15.1508972451,
	b10 = 0.742380924027,
	b11 = 30.789933034,
	b12 = 3.99019417011;
		
	double y, alnorm;
		
	if (z < 0) {
	    upper = 1-upper;
	    z = -z;
	}
    
	if (z <= ltone || (upper==1) && z <= utzero) {
        
	    y = 0.5 * z * z;
        
		if (z > con) {
    		alnorm = b1 * exp(-y) /
		                 (z - b2 +
		                  b3 /
		                  (z + b4 +
		                   b5 /
		                   (z - b6 +
		                    b7 / (z + b8 - b9 / (z + b10 + b11 / (z + b12))))));
		} else {
		    alnorm = 0.5 -
		            z *
		                (a1 - a2 * y / (y + a3 - a4 / (y + a5 + a6 / (y + a7))));
        }
            
	} else {
	    alnorm = 0;
	}

	if (upper==0) {
	    alnorm = 1 - alnorm;
	}
		
	return alnorm;
        
}

double computeAsymptoticProbability(int n,int k,double L){
        
	double Z,numerator, denominator,output; 
		 
	numerator=12.0*(L-0.5)-(3.0*k*n*(n+1.0)*(n+1.0));
	denominator=n*(n+1)*sqrt(k*(n-1.0));
		
	Z=numerator/denominator;

    	output = getTipifiedProbability(Z, 1);
    
	return output;

}

void computeLStatistic(int algA, int algB){

	int j;

	L [algA][algB]=0.0;
		
	for(j=0;j<nCuts;j++){
		L[algA][algB]+=(sumRanks[algA][algB][j]*(j+1.0));
	}
}

void computePValue(int algA, int algB){

	pValue[algA][algB]=computeAsymptoticProbability(nCuts,nFunc,L[algA][algB]);
			
}

void computeLStatisticCorrected(int algA, int algB){

	int j;

	LCorrected [algA][algB]=0.0;
		
	for(j=0;j<nCuts;j++){
		LCorrected[algA][algB]+=(sumRanksCorrected[algA][algB][j]*(j+1.0));
	}
}

void computePValueCorrected(int algA, int algB){

	pValueCorrected[algA][algB]=computeAsymptoticProbability(nCuts,nFunc, LCorrected[algA][algB]);
			
}

void computeRanksCorrected(int algA, int algB){
		
	double rank,min;
	double newRank;
	int count;
	int i,j,k;

	for(i=0;i<nFunc;i++){
		for(j=0;j<nCuts;j++){		
			ranks[algA][algB][i][j]=-1;
		}
	}

	//check whether one of the algorithms has finished before the end
 		
	int searchFinishes [MAX_FUNC];

	for(i=0;i<nFunc;i++){
		searchFinishes[i]=UNDEFINED;
	}
		
	for(i=0;i<nFunc;i++){	
		for(j=0;j<nCuts && searchFinishes[i]==UNDEFINED;j++){
			
			if((results[algA][i][j]==0)&&(results[algB][i][j]==0)){
				searchFinishes[i]=0;
			}
			if((results[algA][i][j]!=0)&&(results[algB][i][j]==0)){
				searchFinishes[i]=-1*j;
			}
			if((results[algA][i][j]==0)&&(results[algB][i][j]!=0)){
				searchFinishes[i]=j;
			}
		}
			
		if(searchFinishes[i]==UNDEFINED){
			searchFinishes[i]=0;
		}
			
	}
		

	for(i=0;i<nFunc;i++){
			
		if(searchFinishes[i]==0){

			rank=1.0;

			do{
				min=DBL_MAX;
				count=0;

				for(j=0;j<nCuts;j++){
					if((ranks[algA][algB][i][j]==-1.0)&&(differences[i][j]==min)){
						count+=1;
					}
					if((ranks[algA][algB][i][j]==-1.0)&&(differences[i][j]<min)){
						min=differences[i][j];
						count=1;
					}
				}

				if(count==1){
					newRank=rank;
				}
				else{
					newRank=0.0;
					for(k=0;k<count;k++){
						newRank+=(rank+k);
					}

					newRank/=(double)count;
				}
				
				for(j=0;j<nCuts;j++){
				
					if(differences[i][j]==min){
						ranks[algA][algB][i][j]=newRank;
					}
				}
				
				rank+=count;

			}while(rank<=nCuts);

		}

		if(searchFinishes[i]>0){
			
			for(j=searchFinishes[i]-1;j<nCuts;j++){
				ranks[algA][algB][i][j]=j+1;
			}
			
			rank=1.0;
				
			do{
				min=DBL_MAX;
				count=0;

				for(j=0;j<searchFinishes[i]-1;j++){

					if((ranks[algA][algB][i][j]==-1.0)&&(differences[i][j]==min)){
						count+=1;
					}

					if((ranks[algA][algB][i][j]==-1.0)&&(differences[i][j]<min)){
						min=differences[i][j];
						count=1;
					}
				}

				if(count==1){
					newRank=rank;
				}
				else{
					newRank=0.0;

					for(k=0;k<count;k++){
						newRank+=(rank+k);
					}

					newRank/=(double)count;
				}
					
				for(j=0;j<searchFinishes[i]-1;j++){
					
					if(differences[i][j]==min){
						ranks[algA][algB][i][j]=newRank;
					}
				}
					
				rank+=count;
			}while(rank<=searchFinishes[i]-1);
			
		}
			
		if(searchFinishes[i]<0){
			
			searchFinishes[i]*=-1;
			rank = nCuts - searchFinishes[i] + 1;

			for(j=searchFinishes[i]-1;j<nCuts;j++){
				ranks[algA][algB][i][j]=rank;
				rank--;
			}
			
			rank = nCuts - searchFinishes[i] + 2;
				
			do{
				min=DBL_MAX;
				count=0;

				for(j=0;j<searchFinishes[i];j++){
					if((ranks[algA][algB][i][j]==-1.0)&&(differences[i][j]==min)){
						count+=1;
					}
					if((ranks[algA][algB][i][j]==-1.0)&&(differences[i][j]<min)){
						min=differences[i][j];
						count=1;
					}
				}

				if(count==1){
					newRank=rank;
				}
				else{
					newRank=0.0;
					for(k=0;k<count;k++){
						newRank+=(rank+k);
					}

					newRank/=(double)count;
				}
					
				for(j=0;j<searchFinishes[i];j++){

					if(differences[i][j]==min){
						ranks[algA][algB][i][j]=newRank;
					}
				}
					
				rank+=count;

			}while(rank<=nCuts);
		}
	}


	for(j=0;j<nCuts;j++){
		sumRanksCorrected[algA][algB][j]=0;
	}

	for(i=0;i<nFunc;i++){
		for(j=0;j<nCuts;j++){
			sumRanksCorrected[algA][algB][j]+=ranks[algA][algB][i][j];
		}
	}

}


void computeRanks(int algA, int algB){
		
	double rank,min;
	double newRank;
	int count;
	int i,j,k;

	for(i=0;i<nFunc;i++){
		for(j=0;j<nCuts;j++){		
			ranks[algA][algB][i][j]=-1;
		}
	}
	
	for(i=0;i<nFunc;i++){

		rank=1.0;

		do{
			min=DBL_MAX;
			count=0;

			for(j=0;j<nCuts;j++){
				if((ranks[algA][algB][i][j]==-1.0)&&(differences[i][j]==min)){
					count+=1;
				}
				if((ranks[algA][algB][i][j]==-1.0)&&(differences[i][j]<min)){
					min=differences[i][j];
					count=1;
				}
			}

			if(count==1){
				newRank=rank;
			}
			else{
				newRank=0.0;
				for(k=0;k<count;k++){
					newRank+=(rank+k);
				}

				newRank/=(double)count;
			}
				
			for(j=0;j<nCuts;j++){
				
				if(differences[i][j]==min){
					ranks[algA][algB][i][j]=newRank;
				}
			}
				
			rank+=count;

		}while(rank<=nCuts);
				
	}
	
	for(j=0;j<nCuts;j++){
		sumRanks[algA][algB][j]=0;
	}

	for(i=0;i<nFunc;i++){
		for(j=0;j<nCuts;j++){
			sumRanks[algA][algB][j]+=ranks[algA][algB][i][j];
		}
	}

}


/*********************
 * Main
 * *******************/  
int main(int argc, char **argv){

	int i,j,k;

	/*********************
	 * Read parameters
	 * *******************/
     
	if(argc < 6){
		printf("Error: Wrong parameters\n");
		printf("Usage: Page <nAlgorithms> <nFunctions> <nCutpoints> [<Alg1ResultsFile> <Alg2ResultsFile>...]\n");
		return -1;
	}

	nAlgs=atoi(argv[1]);
	nFunc=atoi(argv[2]);
	nCuts=atoi(argv[3]);

	if((nAlgs < 2)||(nAlgs > MAX_ALGS)){
		printf("Error: Please use [2,%d] algorithms\n",MAX_ALGS);
		printf("Usage: Page <nAlgorithms> <nFunctions> <nCutpoints> [<Alg1ResultsFile> <Alg2ResultsFile>...]\n");
		return -1;
	}

	if((nFunc < 2)||(nFunc > MAX_FUNC)){
		printf("Error: Please use [2,%d] functions\n",MAX_FUNC);
		printf("Usage: Page <nAlgorithms> <nFunctions> <nCutpoints> [<Alg1ResultsFile> <Alg2ResultsFile>...]\n");
		return -1;
	}

	if((nCuts < 2)||(nCuts > MAX_CUTS)){
		printf("Error: Please use [2,%d] cutpoints\n",MAX_CUTS);
		printf("Usage: Page <nAlgorithms> <nFunctions> <nCutpoints> [<Alg1ResultsFile> <Alg2ResultsFile>...]\n");
		return -1;
	}
    
	/*********************
	 * Load data
	 * *******************/

	names = malloc(nAlgs * sizeof(char*));

	//read name of files
	for(i=0;i<nAlgs;i++){

		names[i] = malloc((strlen(argv[i+4])+1) * sizeof(char));
        	strcpy(names[i], argv[i+4]);

	}
    
	//read file's contents
	for(i=0;i<nAlgs;i++){
    
		FILE *fstream = fopen(names[i],"r");
        
		if(fstream == NULL){
			printf("\n file opening failed ");
			return -1 ;
		}
        
		j=0;
    
		while((line=fgets(buffer,sizeof(buffer),fstream))!=NULL){

			record = strtok(line,",");
	
			double value = atof(record);
			results[i][j][0] = value;

                	for(k=1;k<nCuts;k++){

				record = strtok(NULL,",");		
				value = atof(record);
				results[i][j][k] = value;
                		
            		}

            		j++;
        	}

		fclose(fstream);
	}

	/*********************
	 * Pairwise Page test
	 * *******************/

	for(i=0;i<nAlgs;i++){

		for(j=0;j<nAlgs;j++){

			if(i!=j){
				
				computeDifferences(i,j);
				computeRanks(i,j);
				computeLStatistic(i,j);
				computePValue(i,j);

			}
		}
	}


	for(i=0;i<nAlgs;i++){
		for(j=0;j<nAlgs;j++){

			if(i!=j){
		
				computeDifferences(i,j);
				computeRanksCorrected(i,j);
				computeLStatisticCorrected(i,j);
				computePValueCorrected(i,j);

			}
		}
	}


	/*********************
	 * Print results
	 * *******************/

	FILE * ofl = fopen("output.tex", "w");

	if (ofl == NULL) {
  		printf("\n file writing failed ");
	  	return -1;
	}

	FILE * oft = fopen("output.txt", "w");

	if (oft == NULL) {
  		printf("\n file writing failed ");
	  	return -1;
	}

	//header, LaTeX output

	fprintf(ofl,"\\documentclass[a4paper,10pt]{article}\n");
	fprintf(ofl,"\\title{Results of the Page test for convergence analysis}\n");
	fprintf(ofl,"\\author{}\n");
	fprintf(ofl,"\\date{\\today}\n");
	fprintf(ofl,"\\begin{document}\n");
	fprintf(ofl,"\\oddsidemargin 0in \\topmargin 0in");
	fprintf(ofl,"\\maketitle\n");

	fprintf(ofl,"\\noindent Results are provided for the basic and the corrected versions of the Page Test\n");

	fprintf(ofl,"\\begin{itemize}\n");

	fprintf(ofl,"\\item Ranks are computed from 1 (smaller differences) to N (larger differences).\n");
    
	fprintf(ofl,"\\item Ties are handled using the midrank method (i.e. a tie in ranks 2 and 3 will be represented as 2.5, 2.5).\n");
    
	fprintf(ofl,"\\item The L statistic is computed as\n"); 
    
	fprintf(ofl,"\\begin{equation}\n");
	fprintf(ofl," L = \\sum_{j=1}^{n}Y_{j}R_{j}\n");
	fprintf(ofl,"\\end{equation}\n");
    
	fprintf(ofl,"\\noindent where $Y_{j}$ is the is the hypothetical ranking of the \\textit{jth} treatment (cutpoint, in this case) and $R_{j}$ is the average ranking computed for this treatment.\n");
    
	fprintf(ofl,"\\item The asymptotic $p$-values are computed using the statistic\n");
    
	fprintf(ofl,"\\begin{equation}\n");
	fprintf(ofl,"Z = \\frac{12(L-0.5)-3kn(n+1)^{2}}{n(n+1)\\sqrt{k(n-1)}}\n");
	fprintf(ofl,"\\end{equation}\n");
    
	fprintf(ofl,"\\noindent which includes a continuity correction. This statistic is approximately standard normal and the appropriate rejection region is right tail (Gibbons,2010).\n");
	fprintf(ofl,"\\item Tables containing exact $p$-values for small samples can be found in (Page,1963).\n");
    
	fprintf(ofl,"\\item The corrected version modifies the computation of the rankings so it trully acknowledges those cases in which an algorithm reaches the optimum before the end. See the test description for more details.\n");

	fprintf(ofl,"\\end{itemize}\n");

	fprintf(ofl,"\\section*{References}\n");

	fprintf(ofl,"* J.D. Gibbons, S. Chakraborti (2010), Nonparametric Statistical Inference, 5th edition, Chapman \\& Hall.\n");

	fprintf(ofl,"\n\n\\vspace{4 mm}\n\n");

	fprintf(ofl,"\\noindent * E.P. Page (1963), Ordered hypotheses for multiple treatments: A significance test for linear ranks, Journal of the American Statistical Association, 58, 216--230.\n");

	fprintf(ofl,"\\clearpage\n");

	//header, plain text output

	fprintf(oft,"****************************************************\n");
	fprintf(oft,"Results of the Page test for convergence analysis \n");
	fprintf(oft,"****************************************************\n\n\n");

	fprintf(oft,"Results are provided for the basic and the corrected versions of the Page Test:\n\n");

	fprintf(oft,"- Ranks are computed from 1 (smaller differences) to N (larger differences).\n");
    
	fprintf(oft,"- Ties are handled using the midrank method (i.e. a tie in ranks 2 and 3 will be represented as 2.5, 2.5).\n");
    
	fprintf(oft,"- The L statistic is computed as\n\n"); 

	fprintf(oft,"         L = sum_{j=1}^{n}Y_{j}R_{j}\n\n");
	    	
	fprintf(oft,"  where $Y_{j}$ is the is the hypothetical ranking of the \\textit{jth} treatment (cutpoint, in this case) and $R_{j}$ is the average ranking computed for this treatment.\n");
    
	fprintf(oft,"- The asymptotic p-values are computed using the statistic\n\n");
    
	fprintf(oft,"        Z = frac{12(L-0.5)-3kn(n+1)^{2}}{n(n+1)sqrt{k(n-1)}}\n\n");

	fprintf(oft,"  which includes a continuity correction. This statistic is approximately standard normal and the appropriate rejection region is right tail (Gibbons,2010).\n");
	fprintf(oft,"- Tables containing exact $p$-values for small samples can be found in (Page,1963).\n");
    
	fprintf(oft,"- The corrected version modifies the computation of the rankings so it trully acknowledges those cases in which an algorithm reaches the optimum before the end. See the test description for more details.\n");

	fprintf(oft,"\nReferences:\n\n");

	fprintf(oft,"* J.D. Gibbons, S. Chakraborti (2010), Nonparametric Statistical Inference, 5th edition, Chapman & Hall.\n");

	fprintf(oft,"* E.P. Page (1963), Ordered hypotheses for multiple treatments: A significance test for linear ranks, Journal of the American Statistical Association, 58, 216--230.\n");

	//results, LaTeX output

	fprintf(ofl,"\\section*{Results obtained (Basic version)}\n");
	fprintf(ofl,"\\begin{table}[h]\n");
	fprintf(ofl,"\\begin{center}\n");
	fprintf(ofl,"\\begin{tabular}{|l|c|r|}\n");
	fprintf(ofl,"\\hline\n");
	fprintf(ofl,"Algorithms & L Statistic & $p$-value \\\\ \n");
	fprintf(ofl,"\\hline\n");
 		
	for(i=0;i<nAlgs;i++){
		for(j=0;j<nAlgs;j++){

			if(i!=j){
				fprintf(ofl,"%s VS %s & %.1f & %.5f \\\\ \n",names[i],names[j],L[i][j],pValue [i][j]);
			}

		}

		fprintf(ofl,"\\hline\n");
	}

 	fprintf(ofl,"\\end{tabular}\n");
	fprintf(ofl,"\\end{center}\n");
	fprintf(ofl,"\\end{table}\n");

	fprintf(ofl,"\\clearpage\n");

	fprintf(ofl,"\\section*{Sum of ranks (Basic version)}\n");

	fprintf(ofl,"\\begin{table}[h]\n");
	fprintf(ofl,"\\begin{center}\n");
	fprintf(ofl,"\\begin{tabular}{|l|"); 

	for(k=0;k<nCuts;k++){
		fprintf(ofl,"c|");
	}
	
	fprintf(ofl,"}\n");

	fprintf(ofl,"\\hline\n");
	fprintf(ofl,"Algorithms & "); 

	for(k=0;k<nCuts-1;k++){
		fprintf(ofl,"C%d & ",k+1);
	}

	fprintf(ofl,"C%d \\\\ \n",nCuts);

	fprintf(ofl,"\\hline\n");
 		
	for(i=0;i<nAlgs;i++){
		for(j=0;j<nAlgs;j++){	
			if(i!=j){

				fprintf(ofl,"%s VS %s & ",names[i],names[j]);

				for(k=0;k<nCuts-1;k++){
					fprintf(ofl,"%.1f & ",sumRanks[i][j][k]);
				}

				fprintf(ofl,"%.1f \\\\",sumRanks[i][j][nCuts-1]);	
			}

		}

		fprintf(ofl,"\\hline\n");
	}

 	fprintf(ofl,"\\end{tabular}\n");
	fprintf(ofl,"\\end{center}\n");
	fprintf(ofl,"\\end{table}\n");

	fprintf(ofl,"\\clearpage\n");

	//results, plain text output

	fprintf(oft,"\n\n****************************************************\n");
	fprintf(oft,"Results obtained (Basic version)\n");
	fprintf(oft,"****************************************************\n\n\n");
 	fprintf(oft,"Algorithms\tL-Statistic\tp-Value\n\n");

	for(i=0;i<nAlgs;i++){
		for(j=0;j<nAlgs;j++){

			if(i!=j){
				fprintf(oft,"%s VS %s\t%.1f\t%.5f\n",names[i],names[j],L[i][j],pValue[i][j]);
			}
		}
	}
	
	fprintf(oft,"\n\n");

	fprintf(oft,"\n****************************************************\n");
	fprintf(oft,"Sum of ranks (Basic version)\n");
	fprintf(oft,"****************************************************\n\n\n");

    	fprintf(oft,"Algorithms\t"); 

	for(k=0;k<nCuts-1;k++){
		fprintf(oft,"C%d\t",(k+1));
	}
	
	fprintf(oft,"C%d\n",nCuts);

    	fprintf(oft,"\n");
 		
	for(i=0;i<nAlgs;i++){
		for(j=0;j<nAlgs;j++){	
			if(i!=j){

				fprintf(oft,"%s VS %s\t",names[i],names[j]);

				for(k=0;k<nCuts-1;k++){
					fprintf(oft,"%.1f\t",sumRanks[i][j][k]);
				}

				fprintf(oft,"%.1f\t",sumRanks[i][j][nCuts-1]);	
				fprintf(oft,"\n");
			}
			
		}

		fprintf(oft,"\n");
	}

	fprintf(oft,"\n\n");

	//alternative results, LaTeX output

	fprintf(ofl,"\\section*{Results obtained (Alternative version)}\n");
	fprintf(ofl,"\\begin{table}[h]\n");
	fprintf(ofl,"\\begin{center}\n");
	fprintf(ofl,"\\begin{tabular}{|l|c|r|}\n");
	fprintf(ofl,"\\hline\n");
	fprintf(ofl,"Algorithms & L Statistic & $p$-value \\\\ \n");
	fprintf(ofl,"\\hline\n");
 		
	for(i=0;i<nAlgs;i++){
		for(j=0;j<nAlgs;j++){

			if(i!=j){
				fprintf(ofl,"%s VS %s & %.1f & %.5f \\\\ \n",names[i],names[j],L[i][j],pValueCorrected [i][j]);
			}

		}

		fprintf(ofl,"\\hline\n");
	}

 	fprintf(ofl,"\\end{tabular}\n");
	fprintf(ofl,"\\end{center}\n");
	fprintf(ofl,"\\end{table}\n");

	fprintf(ofl,"\\clearpage\n");

	fprintf(ofl,"\\section*{Sum of ranks (Alternative version)}\n");

	fprintf(ofl,"\\begin{table}[h]\n");
	fprintf(ofl,"\\begin{center}\n");
	fprintf(ofl,"\\begin{tabular}{|l|"); 

	for(k=0;k<nCuts;k++){
		fprintf(ofl,"c|");
	}
	
	fprintf(ofl,"}\n");

	fprintf(ofl,"\\hline\n");
	fprintf(ofl,"Algorithms & "); 

	for(k=0;k<nCuts-1;k++){
		fprintf(ofl,"C%d & ",k+1);
	}

	fprintf(ofl,"C%d \\\\ \n",nCuts);

	fprintf(ofl,"\\hline\n");
 		
	for(i=0;i<nAlgs;i++){
		for(j=0;j<nAlgs;j++){	
			if(i!=j){

				fprintf(ofl,"%s VS %s & ",names[i],names[j]);

				for(k=0;k<nCuts-1;k++){
					fprintf(ofl,"%.1f & ",sumRanksCorrected[i][j][k]);
				}

				fprintf(ofl,"%.1f \\\\",sumRanksCorrected[i][j][nCuts-1]);	
			}

		}

		fprintf(ofl,"\\hline\n");
	}

 	fprintf(ofl,"\\end{tabular}\n");
	fprintf(ofl,"\\end{center}\n");
	fprintf(ofl,"\\end{table}\n");

	fprintf(ofl,"\\clearpage\n");

	//alternative results, plain text output

	fprintf(oft,"\n\n****************************************************\n");
	fprintf(oft,"Results obtained (Alternative version)\n");
	fprintf(oft,"****************************************************\n\n\n");
 	fprintf(oft,"Algorithms\tL-Statistic\tp-Value\n\n");

	for(i=0;i<nAlgs;i++){
		for(j=0;j<nAlgs;j++){

			if(i!=j){
				fprintf(oft,"%s VS %s\t%.1f\t%.5f\n",names[i],names[j],L[i][j],pValueCorrected[i][j]);
			}
		}
	}
	
	fprintf(oft,"\n\n");

	fprintf(oft,"\n****************************************************\n");
	fprintf(oft,"Sum of ranks (Alternative version)\n");
	fprintf(oft,"****************************************************\n\n\n");

    	fprintf(oft,"Algorithms\t"); 

	for(k=0;k<nCuts-1;k++){
		fprintf(oft,"C%d\t",(k+1));
	}
	
	fprintf(oft,"C%d\n",nCuts);

    	fprintf(oft,"\n");
 		
	for(i=0;i<nAlgs;i++){
		for(j=0;j<nAlgs;j++){	
			if(i!=j){

				fprintf(oft,"%s VS %s\t",names[i],names[j]);

				for(k=0;k<nCuts-1;k++){
					fprintf(oft,"%.1f\t",sumRanksCorrected[i][j][k]);
				}

				fprintf(oft,"%.1f\t",sumRanksCorrected[i][j][nCuts-1]);	
				fprintf(oft,"\n");
			}
			
		}

		fprintf(oft,"\n");
	}

	fprintf(oft,"\n\n");

	// footer, Latex output

	fprintf(ofl,"\\end{document}\n");

	fclose(ofl);
	fclose(oft);



	/*********************
	 * Free memory
	 * *******************/

	for(i=0;i<nAlgs;i++){
		free(names[i]);
	} 
    
	free(names);

	return 0;
}
