/***********************************************************************

	This file is a simple script written in Java to apply the Page's
	trend test to analyze evolutionary search algorithms convergence.
	Results are obtained in LaTeX and plain text format.

	Copyright (C) 2014
	
	J. Derrac (derrac@decsai.ugr.es)

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see http://www.gnu.org/licenses/
  
**********************************************************************/

import java.util.*;
import java.io.*;
import java.text.*;

/*

Main class of the script

*/
public class Page{

	public static final int UNDEFINED = 999;
 
	public static double results [][][];
	public static String names [];
	public static int nAlgs;
	public static int nFunc;
	public static int nCuts;
	public static double ranks [][][][];
	public static double sumRanks [][][];
	public static double L [][];
	public static double pValue [][];
	public static double sumRanksCorrected [][][];
	public static double LCorrected [][];
	public static double pValueCorrected [][];
	public static double differences [][];

	public static DecimalFormat nf1, nf4, nf6; 
	public static DecimalFormatSymbols dfs;

	public static String plainOutput;

	/**
	 * Main method
	 * 
	 */
	public static void main(String[] args) {
		
		String text="";

		setDecimalFormat();
		
		//Read arguments

		if(args.length < 5){
			System.out.println("Error: Wrong parameters");
			System.out.println("Usage: Page <nAlgorithms> <nFunctions> <nCutpoints> [<Alg1ResultsFile> <Alg2ResultsFile>]...");
			System.exit(0);
		}

		nAlgs = Integer.parseInt(args[0]);
		nFunc = Integer.parseInt(args[1]);
		nCuts = Integer.parseInt(args[2]);

		
		//Read average results of each algorithm

		results = new double [nAlgs][nFunc][nCuts];
		names = new String [nAlgs];
		
		for(int i=0;i<nAlgs;i++){
			names[i]= args[i+3];
			text=readFile(names[i]);
			copyResults(i,text);
		}

		ranks = new double [nAlgs][nAlgs][nFunc][nCuts];
		sumRanks = new double [nAlgs][nAlgs][nCuts];
		L = new double [nAlgs][nAlgs];
		pValue = new double [nAlgs][nAlgs];

		sumRanksCorrected = new double [nAlgs][nAlgs][nCuts];
		LCorrected = new double [nAlgs][nAlgs];
		pValueCorrected = new double [nAlgs][nAlgs];

		//Perform the test for each pair of algorithms (original version)
		for(int i=0;i<nAlgs;i++){
			for(int j=0;j<nAlgs;j++){

				if(i!=j){
		
					computeDifferences(i,j);
					
					//original version
				
					computeRanks(i,j);

					computeLStatistic(i,j);

					computePValue(i,j);

				}

			}
		}


		//Perform the test for each pair of algorithms (corrected version)
		for(int i=0;i<nAlgs;i++){
			for(int j=0;j<nAlgs;j++){

				if(i!=j){
				
					computeDifferences(i,j);

					//corrected version
				
					computeRanksCorrected(i,j);

					computeLStatisticCorrected(i,j);

					computePValueCorrected(i,j);

				}

			}
		}

		//Print final results
		text="";

		text+=printHeader();

		text+=printResults();

		text+=printCorrectedResults();

		text+=printFoot();

		plainOutput="";

		plainOutput+=printPlainHeader();

		plainOutput+=printPlainResults();

		plainOutput+=printPlainCorrectedResults();

		try{
			PrintWriter pr = new PrintWriter("output.tex");

			pr.println(text);

			pr.close();

			pr = new PrintWriter("output.txt");

			pr.println(plainOutput);

			pr.close();

		}catch(Exception e){
			e.printStackTrace();
		}
		
	}

	/**
	 * Print the header of the results file
	 * 
	 */
	private static String printPlainHeader(){

		String text="";

		text+="****************************************************\n";
	    	text+="Results of the Page test for convergence analysis \n";
		text+="****************************************************\n\n\n";

	    	text+="Results are provided for the basic and the corrected versions of the Page Test:\n\n";

	    	text+="- Ranks are computed from 1 (smaller differences) to N (larger differences).\n";
    
	    	text+="- Ties are handled using the midrank method (i.e. a tie in ranks 2 and 3 will be represented as 2.5, 2.5).\n";
    
	    	text+="- The L statistic is computed as\n\n"; 

	    	text+="         L = sum_{j=1}^{n}Y_{j}R_{j}\n\n";
	    	
	    	text+="  where $Y_{j}$ is the is the hypothetical ranking of the \\textit{jth} treatment (cutpoint, in this case) and $R_{j}$ is the average ranking computed for this treatment.\n";
    
	    	text+="- The asymptotic p-values are computed using the statistic\n\n";
    
	    	text+="        Z = frac{12(L-0.5)-3kn(n+1)^{2}}{n(n+1)sqrt{k(n-1)}}\n\n";

	    	text+="  which includes a continuity correction. This statistic is approximately standard normal and the appropriate rejection region is right tail (Gibbons,2010).\n";
	    	text+="- Tables containing exact $p$-values for small samples can be found in (Page,1963).\n";
    
	    	text+="- The corrected version modifies the computation of the rankings so it trully acknowledges those cases in which an algorithm reaches the optimum before the end. See the test description for more details.\n";

	    	text+="\nReferences:\n\n";

	    	text+="* J.D. Gibbons, S. Chakraborti (2010), Nonparametric Statistical Inference, 5th edition, Chapman & Hall.\n";

	    	text+="* E.P. Page (1963), Ordered hypotheses for multiple treatments: A significance test for linear ranks, Journal of the American Statistical Association, 58, 216--230.\n";

		return text;
	}

	/**
	 * Print the header of the results file
	 * 
	 */
	private static String printHeader(){

		String text="";

		text+="\\documentclass[a4paper,10pt]{article}\n";
	    	text+="\\title{Results of the Page test for convergence analysis}\n";
	    	text+="\\author{}\n";
	    	text+="\\date{\\today}\n";
	    	text+="\\begin{document}\n";
		text+="\\oddsidemargin 0in \\topmargin 0in";
	    	text+="\\maketitle\n";

	    	text+="\\noindent Results are provided for the basic and the corrected versions of the Page Test\n";

	    	text+="\\begin{itemize}\n";

	    	text+="\\item Ranks are computed from 1 (smaller differences) to N (larger differences).\n";
    
	    	text+="\\item Ties are handled using the midrank method (i.e. a tie in ranks 2 and 3 will be represented as 2.5, 2.5).\n";
    
	    	text+="\\item The L statistic is computed as\n"; 
    
	    	text+="\\begin{equation}\n";
	    	text+=" L = \\sum_{j=1}^{n}Y_{j}R_{j}\n";
	    	text+="\\end{equation}\n";
    
	    	text+="\\noindent where $Y_{j}$ is the is the hypothetical ranking of the \\textit{jth} treatment (cutpoint, in this case) and $R_{j}$ is the average ranking computed for this treatment.\n";
    
	    	text+="\\item The asymptotic $p$-values are computed using the statistic\n";
    
	    	text+="\\begin{equation}\n";
	    	text+="Z = \\frac{12(L-0.5)-3kn(n+1)^{2}}{n(n+1)\\sqrt{k(n-1)}}\n";
	    	text+="\\end{equation}\n";
    
	    	text+="\\noindent which includes a continuity correction. This statistic is approximately standard normal and the appropriate rejection region is right tail (Gibbons,2010).\n";
	    	text+="\\item Tables containing exact $p$-values for small samples can be found in (Page,1963).\n";
    
	    	text+="\\item The corrected version modifies the computation of the rankings so it trully acknowledges those cases in which an algorithm reaches the optimum before the end. See the test description for more details.\n";

	    	text+="\\end{itemize}\n";

	    	text+="\\section*{References}\n";

	    	text+="* J.D. Gibbons, S. Chakraborti (2010), Nonparametric Statistical Inference, 5th edition, Chapman \\& Hall.\n";

	    	text+="\n\n\\vspace{4 mm}\n\n";

	    	text+="\\noindent * E.P. Page (1963), Ordered hypotheses for multiple treatments: A significance test for linear ranks, Journal of the American Statistical Association, 58, 216--230.\n";

		text+="\\clearpage\n";
	    	

		return text;
	}

	/**
	 * Print the foot of the results file
	 * 
	 */
	private static String printFoot(){

		String text="";

		text+="\\end{document}\n";   	

		return text;
	}

	/**
	 * Print the results of the corrected version
	 * 
	 */
	private static String printCorrectedResults(){

		String text="";

		text+="\\section*{Results obtained (Corrected version)}\n";
		text+="\\begin{table}[h]\n";
		text+="\\begin{center}\n";
  		text+="\\begin{tabular}{|l|c|r|}\n";
    		text+="\\hline\n";
    		text+="Algorithms & L Statistic & $p$-value \\\\ \n";
    		text+="\\hline\n";
 		
		for(int i=0;i<nAlgs;i++){
			for(int j=0;j<nAlgs;j++){

				if(i!=j){
					text+=names[i]+" VS "+names[j]+" & "+nf4.format(LCorrected[i][j])+" & "+ nf6.format(pValueCorrected [i][j])+" \\\\ \n";
				}


			}
			text+="\\hline\n";
		}

 		text+="\\end{tabular}\n";
		text+="\\end{center}\n";
		text+="\\end{table}\n";

		text+="\\clearpage\n";

		text+="\\section*{Sum of ranks (Corrected version)}\n";

		text+="\\begin{table}[h]\n";
		text+="\\begin{center}\n";
  		text+="\\begin{tabular}{|l|"; for(int k=0;k<nCuts;k++){text+="c|";};text+="}\n";
    		text+="\\hline\n";
    		text+="Algorithms & "; for(int k=0;k<nCuts-1;k++){text+="C"+(k+1)+" & ";};text+="C"+(nCuts)+" \\\\ \n";
    		text+="\\hline\n";
 		
		for(int i=0;i<nAlgs;i++){
			for(int j=0;j<nAlgs;j++){	
				if(i!=j){
					text+=names[i]+" VS "+names[j]+" & ";
					for(int k=0;k<nCuts-1;k++){
						text+=nf1.format(sumRanksCorrected[i][j][k])+" & ";
					}
					text+=nf1.format(sumRanksCorrected[i][j][nCuts-1])+" \\\\";	
				}


			}
			text+="\\hline\n";
		}

 		text+="\\end{tabular}\n";
		text+="\\end{center}\n";
		text+="\\end{table}\n";

		text+="\\clearpage\n";
	    	

		return text;
	}

	/**
	 * Print the results of the corrected version
	 * 
	 */
	private static String printPlainCorrectedResults(){

		String text="";

		text+="\n\n****************************************************\n";
	    	text+="Results obtained (Alternative version)\n";
		text+="****************************************************\n\n\n";
 		text+="Algorithms\tL-Statistic\tp-Value\n\n";
		for(int i=0;i<nAlgs;i++){
			for(int j=0;j<nAlgs;j++){

				if(i!=j){
					text+=names[i]+" VS "+names[j]+"\t"+nf4.format(LCorrected[i][j])+"\t"+ nf6.format(pValueCorrected [i][j])+"\n";
				}


			}
		}
	
		text+="\n\n";

		text+="\n****************************************************\n";
	    	text+="Sum of ranks (Alternative version)\n";
		text+="****************************************************\n\n\n";

    		text+="Algorithms\t"; for(int k=0;k<nCuts-1;k++){text+="C"+(k+1)+"\t";};text+="C"+(nCuts)+"\n";
    		text+="\n";
 		
		for(int i=0;i<nAlgs;i++){
			for(int j=0;j<nAlgs;j++){	
				if(i!=j){
					text+=names[i]+" VS "+names[j]+"\t";
					for(int k=0;k<nCuts-1;k++){
						text+=nf1.format(sumRanksCorrected[i][j][k])+"\t";
					}
					text+=nf1.format(sumRanksCorrected[i][j][nCuts-1]);	
				}


			}
			text+="\n";
		}

		text+="\n\n";

		return text;

	}

	/**
	 * Print the results of the basic version
	 * 
	 */
	private static String printPlainResults(){

		String text="";

		text+="\n\n****************************************************\n";
	    	text+="Results obtained (Basic version)\n";
		text+="****************************************************\n\n\n";
 		text+="Algorithms\tL-Statistic\tp-Value\n\n";
		for(int i=0;i<nAlgs;i++){
			for(int j=0;j<nAlgs;j++){

				if(i!=j){
					text+=names[i]+" VS "+names[j]+"\t"+nf4.format(L[i][j])+"\t"+ nf6.format(pValue [i][j])+"\n";
				}


			}
		}
	
		text+="\n\n";

		text+="\n****************************************************\n";
	    	text+="Sum of ranks (Basic version)\n";
		text+="****************************************************\n\n\n";


    		text+="Algorithms\t"; for(int k=0;k<nCuts-1;k++){text+="C"+(k+1)+"\t";};text+="C"+(nCuts)+"\n";
    		text+="\n";
 		
		for(int i=0;i<nAlgs;i++){
			for(int j=0;j<nAlgs;j++){	
				if(i!=j){
					text+=names[i]+" VS "+names[j]+"\t";
					for(int k=0;k<nCuts-1;k++){
						text+=nf1.format(sumRanks[i][j][k])+"\t";
					}
					text+=nf1.format(sumRanks[i][j][nCuts-1]);	
				}


			}
			text+="\n";
		}

		text+="\n\n";

		return text;
	}

	/**
	 * Print the results of the basic version
	 * 
	 */
	private static String printResults(){

		String text="";

		text+="\\section*{Results obtained (Basic version)}\n";
		text+="\\begin{table}[h]\n";
		text+="\\begin{center}\n";
  		text+="\\begin{tabular}{|l|c|r|}\n";
    		text+="\\hline\n";
    		text+="Algorithms & L Statistic & $p$-value \\\\ \n";
    		text+="\\hline\n";
 		
		for(int i=0;i<nAlgs;i++){
			for(int j=0;j<nAlgs;j++){

				if(i!=j){
					text+=names[i]+" VS "+names[j]+" & "+nf4.format(L[i][j])+" & "+ nf6.format(pValue [i][j])+" \\\\ \n";
				}


			}
			text+="\\hline\n";
		}
;
 		text+="\\end{tabular}\n";
		text+="\\end{center}\n";
		text+="\\end{table}\n";

		text+="\\clearpage\n";

		text+="\\section*{Sum of ranks (Basic version)}\n";

		text+="\\begin{table}[h]\n";
		text+="\\begin{center}\n";
  		text+="\\begin{tabular}{|l|"; for(int k=0;k<nCuts;k++){text+="c|";};text+="}\n";
    		text+="\\hline\n";
    		text+="Algorithms & "; for(int k=0;k<nCuts-1;k++){text+="C"+(k+1)+" & ";};text+="C"+(nCuts)+" \\\\ \n";
    		text+="\\hline\n";
 		
		for(int i=0;i<nAlgs;i++){
			for(int j=0;j<nAlgs;j++){	
				if(i!=j){
					text+=names[i]+" VS "+names[j]+" & ";
					for(int k=0;k<nCuts-1;k++){
						text+=nf1.format(sumRanks[i][j][k])+" & ";
					}
					text+=nf1.format(sumRanks[i][j][nCuts-1])+" \\\\";	
				}


			}
			text+="\\hline\n";
		}

 		text+="\\end{tabular}\n";
		text+="\\end{center}\n";
		text+="\\end{table}\n";

		text+="\\clearpage\n";
	    	

		return text;
	}

	/**
	 * Format of decimal values
	 * 
	 */
	private static void setDecimalFormat(){

		nf1 = (DecimalFormat) DecimalFormat.getInstance();
		nf1.setMaximumFractionDigits(4);
		nf1.setMinimumFractionDigits(0);
		nf1.setGroupingUsed(false);

		dfs = nf1.getDecimalFormatSymbols();
		dfs.setDecimalSeparator('.');
		nf1.setDecimalFormatSymbols(dfs);

		nf4 = (DecimalFormat) DecimalFormat.getInstance();
		nf4.setMaximumFractionDigits(4);
		nf4.setMinimumFractionDigits(0);
		nf4.setGroupingUsed(false);

		dfs = nf4.getDecimalFormatSymbols();
		dfs.setDecimalSeparator('.');
		nf4.setDecimalFormatSymbols(dfs);

		nf6 = (DecimalFormat) DecimalFormat.getInstance();
		nf6.setMaximumFractionDigits(6);
		nf6.setMinimumFractionDigits(6);
		nf6.setGroupingUsed(false);

		dfs = nf6.getDecimalFormatSymbols();
		dfs.setDecimalSeparator('.');
		nf6.setDecimalFormatSymbols(dfs);

	}

	/**
	 * Computation of L statistic
	 * 
	 */
	private static void computeLStatistic(int algA, int algB){

		L [algA][algB]=0.0;
		
		for(int j=0;j<nCuts;j++){
			L[algA][algB]+=(sumRanks[algA][algB][j]*(j+1.0));
		}
	}

	/**
	 * Computation of asymptotic p-value
	 * 
	 */
	private static void computePValue(int algA, int algB){

		pValue[algA][algB]=computeAsymptoticProbability(nCuts,nFunc, L[algA][algB]);
			
	}

	/**
	 * Computation of L statistic (corrected version)
	 * 
	 */
	private static void computeLStatisticCorrected(int algA, int algB){

		LCorrected [algA][algB]=0.0;
		
		for(int j=0;j<nCuts;j++){
			LCorrected[algA][algB]+=(sumRanksCorrected[algA][algB][j]*(j+1.0));
		}
	}

	/**
	 * Computation of asymptotic p-value (corrected version)
	 * 
	 */
	private static void computePValueCorrected(int algA, int algB){

		pValueCorrected[algA][algB]=computeAsymptoticProbability(nCuts,nFunc, LCorrected[algA][algB]);
			
	}

	/**
	 * Compute ranks for the Page test
	 * (1..n, k rows). Ties are handled by the midrank method
	 */
	private static void computeRanks(int algA, int algB){
		
		double rank,min;
		double newRank;
		int count;

		for(int i=0;i<differences.length;i++){
			Arrays.fill(ranks[algA][algB][i],-1.0);
		}
		
		for(int i=0;i<differences.length;i++){
			
			rank=1.0;
			do{
				min=Double.MAX_VALUE;
				count=0;
				for(int j=0;j<differences[i].length;j++){
					if((ranks[algA][algB][i][j]==-1.0)&&(differences[i][j]==min)){
						count+=1;
					}
					if((ranks[algA][algB][i][j]==-1.0)&&(differences[i][j]<min)){
						min=differences[i][j];
						count=1;
					}
				}

				if(count==1){
					newRank=rank;
				}
				else{
					newRank=0.0;
					for(int k=0;k<count;k++){
						newRank+=(rank+k);
					}

					newRank/=(double)count;
				}
				
				for(int j=0;j<differences[i].length;j++){
				
					if(differences[i][j]==min){
						ranks[algA][algB][i][j]=newRank;
					}
				}
				
				rank+=count;

			}while(rank<=differences[0].length);
				
		}

		Arrays.fill(sumRanks[algA][algB],0);
		for(int i=0;i<nFunc;i++){
			for(int j=0;j<nCuts;j++){
				sumRanks[algA][algB][j]+=ranks[algA][algB][i][j];
			}
		}
		
	}

	/**
	 * Compute ranks for the Page test (corrected version)
	 * (1..n, k rows). Ties are handled by the midrank method
	 */
	private static void computeRanksCorrected(int algA, int algB){
		
		double rank,min;
		double newRank;
		int count;

		for(int i=0;i<differences.length;i++){
			Arrays.fill(ranks[algA][algB][i],-1.0);
		}

		//check whether one of the algorithms has finished before the end
 		
		int [] searchFinishes = computeFinished(algA, algB);

		for(int i=0;i<differences.length;i++){
			
			if(searchFinishes[i]==0){
				rank=1.0;
				do{
					min=Double.MAX_VALUE;
					count=0;
					for(int j=0;j<differences[i].length;j++){
						if((ranks[algA][algB][i][j]==-1.0)&&(differences[i][j]==min)){
							count+=1;
						}
						if((ranks[algA][algB][i][j]==-1.0)&&(differences[i][j]<min)){
							min=differences[i][j];
							count=1;
						}
					}

					if(count==1){
						newRank=rank;
					}
					else{
						newRank=0.0;
						for(int k=0;k<count;k++){
							newRank+=(rank+k);
						}

						newRank/=(double)count;
					}
				
					for(int j=0;j<differences[i].length;j++){
				
						if(differences[i][j]==min){
							ranks[algA][algB][i][j]=newRank;
						}
					}
				
					rank+=count;

				}while(rank<=differences[0].length);
					
			}
			if(searchFinishes[i]>0){
			
				for(int j=searchFinishes[i]-1;j<differences[0].length;j++){
					ranks[algA][algB][i][j]=j+1;
				}
			
				rank=1.0;
				
				do{
					min=Double.MAX_VALUE;
					count=0;
					for(int j=0;j<searchFinishes[i]-1;j++){

						if((ranks[algA][algB][i][j]==-1.0)&&(differences[i][j]==min)){
							count+=1;
						}
						if((ranks[algA][algB][i][j]==-1.0)&&(differences[i][j]<min)){
							min=differences[i][j];
							count=1;
						}
					}

					if(count==1){
						newRank=rank;
					}
					else{
						newRank=0.0;
						for(int k=0;k<count;k++){
							newRank+=(rank+k);
						}

						newRank/=(double)count;
					}
					
					for(int j=0;j<searchFinishes[i]-1;j++){
					
						if(differences[i][j]==min){
							ranks[algA][algB][i][j]=newRank;
						}
					}
					
					rank+=count;
				}while(rank<=searchFinishes[i]-1);
			
			}
			if(searchFinishes[i]<0){
			
				searchFinishes[i]*=-1;
				rank = differences[0].length - searchFinishes[i] + 1;
				for(int j=searchFinishes[i]-1;j<differences[0].length;j++){
					ranks[algA][algB][i][j]=rank;
					rank--;
				}
			
				rank = differences[0].length - searchFinishes[i] + 2;
				
				do{
					min=Double.MAX_VALUE;
					count=0;
					for(int j=0;j<searchFinishes[i];j++){
						if((ranks[algA][algB][i][j]==-1.0)&&(differences[i][j]==min)){
							count+=1;
						}
						if((ranks[algA][algB][i][j]==-1.0)&&(differences[i][j]<min)){
							min=differences[i][j];
							count=1;
						}
					}

					if(count==1){
						newRank=rank;
					}
					else{
						newRank=0.0;
						for(int k=0;k<count;k++){
							newRank+=(rank+k);
						}

						newRank/=(double)count;
					}
					
					for(int j=0;j<searchFinishes[i];j++){

						if(differences[i][j]==min){
							ranks[algA][algB][i][j]=newRank;
						}
					}
					
					rank+=count;

				}while(rank<=differences[0].length);
			}
		}

		Arrays.fill(sumRanksCorrected[algA][algB],0);
		for(int i=0;i<nFunc;i++){
			for(int j=0;j<nCuts;j++){
				sumRanksCorrected[algA][algB][j]+=ranks[algA][algB][i][j];
			}
		}
		
	}

	/**
	 * Compute at which cutpoints each algorithm has finished (per function)
	 *
	 */
	private static int [] computeFinished(int algA, int algB){
	
		int [] array;

		array=new int [nFunc];
		
		Arrays.fill(array,UNDEFINED);
		
		for(int i=0;i<nFunc;i++){	
			for(int j=0;j<nCuts && array[i]==UNDEFINED;j++){
			
				if((results[algA][i][j]==0)&&(results[algB][i][j]==0)){
					array[i]=0;
				}
				if((results[algA][i][j]!=0)&&(results[algB][i][j]==0)){
					array[i]=-1*j;
				}
				if((results[algA][i][j]==0)&&(results[algB][i][j]!=0)){
					array[i]=j;
				}
			}
			
			if(array[i]==UNDEFINED){
				array[i]=0;
			}
			
		}
		
		return array;
	}

	/**
	 * Compute diferences between two algorithms' average results
	 *
	 */
	private static void computeDifferences(int algA, int algB){

		differences= new double [nFunc][nCuts];
		
		for(int i=0;i<nFunc;i++){		
			for(int j=0;j<nCuts;j++){
				differences[i][j]=results[algA][i][j]-results[algB][i][j];
			}
		}
	}

	/**
	 * Fill results matrixes
	 *
	 */
	private static void copyResults(int alg, String text){

		StringTokenizer lines, tokens;
		String line, token;
		

		lines = new StringTokenizer (text,"\n\r");

		for(int i=0;i<nFunc;i++){
			line=lines.nextToken();
			tokens = new StringTokenizer(line,",\t");

			for(int j=0;j<nCuts;j++){
				token = tokens.nextToken();
				results[alg][i][j]=Double.parseDouble(token);
			}
			
		}
	}

	/**
	 * Read a file in text mode
	 *
	 */
	private static String readFile(String name){

		String line, text;
		BufferedReader br;

		text="";

		try {
		
			br = new BufferedReader(new FileReader(name));

			do{
				line = br.readLine();
				text+=line;
				text+="\n";
			}while (line != null); 

			br.close();

		} catch(Exception e){
			e.printStackTrace();
		}

		return text;
	}

	/**
	 * Computes asymptotic p-value of the Page distribution.
	 * 
	 * @param n number of columns
	 * @param k number of rows
	 * @param L Page statistic
	 * @return p-value computed
	 */
	public static double computeAsymptoticProbability(int n,int k,double L){
		
		double Z,numerator, denominator; 
		
		numerator=12.0*(L-0.5)-(3.0*k*n*(n+1.0)*(n+1.0));
		denominator=n*(n+1)*Math.sqrt(k*(n-1.0));
		
		Z=numerator/denominator;

		return getTipifiedProbability(Z, true);

	}
	
	/**
     	* Computes cumulative N(0,1) distribution. 
     	* 
     	* Based on Algorithm AS66 Applied Statistics (1973) vol 22 no.3
     	* 
     	* @param z x value
     	* @param upper A boolean value, if true the integral is evaluated from z to infinity, from minus infinity to z otherwise
     	* @return The value of the cumulative N(0,1) distribution for z
     	*/  
    	public static double getTipifiedProbability(double z, boolean upper) {
		//Algorithm AS 66: "The Normal Integral"
		//Applied Statistics
	    	double ltone = 7.0,
		utzero = 18.66,
		con = 1.28,
		a1 = 0.398942280444,
		a2 = 0.399903438504,
		a3 = 5.75885480458,
		a4 = 29.8213557808,
		a5 = 2.62433121679,
		a6 = 48.6959930692,
		a7 = 5.92885724438,
		b1 = 0.398942280385,
		b2 = 3.8052e-8,
		b3 = 1.00000615302,
		b4 = 3.98064794e-4,
		b5 = 1.986153813664,
		b6 = 0.151679116635,
		b7 = 5.29330324926,
		b8 = 4.8385912808,
		b9 = 15.1508972451,
		b10 = 0.742380924027,
		b11 = 30.789933034,
		b12 = 3.99019417011;
		
		double y, alnorm;
		
		if (z < 0) {
		    upper = !upper;
		    z = -z;
		}
		if (z <= ltone || upper && z <= utzero) {
		    y = 0.5 * z * z;
		    if (z > con) {
		        alnorm = b1 * Math.exp( -y) /
		                 (z - b2 +
		                  b3 /
		                  (z + b4 +
		                   b5 /
		                   (z - b6 +
		                    b7 / (z + b8 - b9 / (z + b10 + b11 / (z + b12))))));
		    } else {
		        alnorm = 0.5 -
		                 z *
		                 (a1 - a2 * y / (y + a3 - a4 / (y + a5 + a6 / (y + a7))));
		    }
		} else {
		    alnorm = 0;
		}
		if (!upper) {
		    alnorm = 1 - alnorm;
		}
		
		return (alnorm);
        
    	}

	
}

